import logging
# import re
import json
import functools
from typing import Optional, Tuple
logger = logging.getLogger('sigmund')


class BaseTool:
    """A base class for tools that process an AI reply."""
    
    def __init__(self, sigmund):
        self._sigmund = sigmund
        
    @property
    def tool_spec(self):
        """The tool spec should corresond to the OpenAI specification for
        function tools.
        """
        return {
            "name": self.__class__.__name__,
            "description": self.__doc__,
            "parameters": {
                "type": "object",
                "properties": self.arguments,
                "required": self.required_arguments,
            }
        }
        
    @property
    def name(self):
        return self.__class__.__name__
    
    def bind(self, args: str, message_prefix: str | None = None) -> callable:
        """Returns a callable that corresponds to a tool function called with
        a string of arguments, which should be in JSON format. The callable 
        itself returns a (progress_message: str, result_message, result: dict,
        needs_feedback: bool) tuple, where messages are informative text message 
        as generated by the tool, result is a dict with name, args, and content 
        keys that correspond to the name and arguments of the function and the 
        result of the tool call. needs_feedback indicates whether the model 
        should be called again to provide feedback based on the tool result.
        """
        def bound_tool_function():
            try:
                kwargs = json.loads(args)
                # import pprint
                # print('*** arguments')
                # pprint.pprint(kwargs)
                # print('***')
                tool_response = self(**kwargs)
            except Exception as e:
                message = 'Failed to run tool'
                result = f'The following error occurred while trying to run tool:\n\n{e}'
                needs_feedback = True
                language = 'markdown'
            else:
                if len(tool_response) == 3:
                    message, result, needs_feedback = tool_response
                    language = 'markdown'
                elif len(tool_response) == 4:
                    message, result, language, needs_feedback = tool_response
                else:
                    raise ValueError(f'Invalid tool response: {tool_response}')
            result = {'name': self.name,
                      'args': args,
                      'content': result}
            if message_prefix is not None:
                message = message_prefix + message
            return message, result, language, needs_feedback
        return bound_tool_function
        
    def __call__(self) -> Tuple[str, Optional[str], bool]:
        """Should be implemented in a tool with additional arguments that
        match tool specification.
        """
        raise NotImplementedError()



import logging
logger = logging.getLogger('sigmund')


db_initialized = False


def predict(prompt, model: str, json: bool = False,
            max_json_retry: int = 5) -> str | dict:
    """Provides a static interface to predict a single response.

    Parameters
    ----------
    prompt : str
        The prompt to predict a response for.
    model : str
        The model to use for prediction.
    json : bool, optional
        Whether to return the response as JSON, by default False.
    max_json_retry : int, optional
        The maximum number of retries to attempt if the response is not valid
        JSON, by default 5.

    Returns
    -------
    str | dict
        The predicted response. If json=True, the response is a dict, otherwise
        it is a str.
    """
    from .database.models import init_db
    from .model import model as sigmund_model
    global db_initialized
    if not db_initialized:
        init_db()
        db_initialized = True
    model = sigmund_model(None, model)
    if not json:
        return model.predict(prompt)
    import json
    from .model import BaseModel
    model.json_mode = True
    for _ in range(max_json_retry):
        reply = model.predict(prompt)
        reply = BaseModel.extract_thinking_block(reply)[0]
        if reply.startswith('```json'):
            reply = reply[7:]
        if reply.startswith('```'):
            reply = reply[3:]
        if reply.endswith('```'):
            reply = reply[:-3]
        try:
            return json.loads(reply)
        except Exception as e:
            logger.warning(f'failed to predict JSON: {e}')
            continue
    raise ValueError(f'Failed after {max_json_retry} attempts')



import json
import base64
from pathlib import Path
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from flask import redirect, url_for, session, Blueprint
from flask_login import login_user, current_user, logout_user, UserMixin
from .. import config
from .. import utils
from .. import process_sigmund_message
from ..forms import LoginForm
from ..sigmund import Sigmund
import logging
logger = logging.getLogger('sigmund')
app_blueprint = Blueprint('app', __name__)


class User(UserMixin):
    def __init__(self, username):
        self.id = username
        logger.info(f'initializing user id: {self.id}')


def get_sigmund(db_cache='default'):
    config.db_cache = db_cache
    return Sigmund(user_id=current_user.get_id(), persistent=True,
                   encryption_key=session['encryption_key'])
    

def chat_page():
    sigmund = get_sigmund()
    if config.subscription_required and \
            not sigmund.database.check_subscription():
        return redirect(url_for('subscribe.subscribe'), code=303)
    html_content = ''
    previous_timestamp = None
    previous_answer_model = None
    workspace_content = ''
    workspace_language = 'markdown'
    for role, message, metadata in sigmund.messages.visible_messages():
        message_id = metadata.get('message_id', 0)
        delete_button = f'<button class="message-delete" onclick="deleteMessage(\'{message_id}\')"><i class="fas fa-trash"></i></button>'
        if role == 'assistant':
            html_body = utils.md(
                f'{config.ai_name}: {process_sigmund_message.process_ai_message(message)}')
            html_class = 'message-ai'
        else:
            html_body = '<p>' + utils.clean(f'You: {message}', 
                                            escape_html=True,
                                            render=False) + '</p>'
            html_class = 'message-user'
        workspace_content = metadata.get('workspace_content', None)
        workspace_language = metadata.get('workspace_language', None)
        if workspace_content:
            workspace_div = f'''<div class="message-workspace" id="message-workspace-{message_id}">
                <a href="#" onclick="loadMessageWorkspace('{message_id}')">Load workspace</a>
                <pre class="workspace-content">{process_sigmund_message.process_workspace_content(workspace_content)}</pre>
                <div class="workspace-language">{workspace_language}</div>
                </div>'''
        else:
            workspace_div = ''
        if 'sources' in metadata:
            sources_div = '<div class="message-sources">'
            sources = json.loads(metadata['sources'])
            urls = {source['url'] for source in sources if source['url']}
            for url in urls:
                sources_div += f'<a href="{url}">{url}</a><br />'
            sources_div += '</div>'
        else:
            sources_div = ''
        if previous_timestamp != metadata['timestamp']:
            previous_timestamp = metadata['timestamp']
            timestamp_div = f'<div class="message-timestamp">{metadata["timestamp"]}</div>'
        else:
            timestamp_div = ''
        if role == 'assistant' and previous_answer_model != metadata['answer_model']:
            previous_answer_model = metadata['answer_model']
            answer_model_div = f'<div class="message-answer-model">{metadata["answer_model"]}</div>'
        else:
            answer_model_div = ''
            
        html_content += f'<div class="message {html_class}" data-message-id="{message_id}">{delete_button}{html_body}{workspace_div}{timestamp_div}{answer_model_div}{sources_div}</div>'
    # The user's settings are the defaults updated with the user-specific 
    # settings from the database
    settings = config.settings_default.copy()
    settings.update(sigmund.database.list_settings())
    if workspace_content is None:
        workspace_content = ''
        workspace_language = 'markdown'
    username = sigmund.user_id
    if '(google)::' in username:
        username = username.split('(google)::')[0] + '(Google)'
    return utils.render('chat.html', message_history=html_content,
                        subscription_required=config.subscription_required,
                        username=username,
                        settings=json.dumps(settings),
                        workspace_content=workspace_content,
                        workspace_language=workspace_language)


def login_handler(form, failed=False):
    if form.validate_on_submit():
        username = form.username.data.strip().lower()
        password = form.password.data.strip()
        username = config.validate_user(username, password)
        if username is None:
            logger.info(f'log-in failed ({username})')
            return redirect('/login_failed')
        logger.info(f'log-in successful ({username})')
        kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                         length=32,
                         salt=config.encryption_salt,
                         iterations=100000,
                         backend=default_backend())
        session['encryption_key'] = base64.urlsafe_b64encode(
            kdf.derive(password.encode()))
        user = User(username)
        login_user(user)
        logger.info('initializing encryption key')
        return redirect('/')
    html_content = utils.render('welcome.html')
    return utils.render('chat.html', message_history=html_content,
                        subscription_required=config.subscription_required,
                        form=form,
                        need_login=True,
                        login_failed=failed,
                        username='unknown friend',
                        settings='{}',
                        workspace_content='',
                        workspace_language='')
    

@app_blueprint.route('/terms')
def terms():
    return utils.render(
        'info-page.html',
        content=utils.md(Path('sigmund/static/terms.md').read_text()))


@app_blueprint.route('/login', methods=['GET', 'POST'])
def login():
    if current_user.is_authenticated:
        return redirect(url_for('app.chat'))
    return login_handler(LoginForm())


@app_blueprint.route('/login_failed', methods=['GET', 'POST'])
def login_failed():
    return login_handler(LoginForm(), failed=True)


@app_blueprint.route('/logout')
def logout():
    logout_user()
    return redirect(url_for('app.login'))


@app_blueprint.route('/')
def main():
    if current_user.is_authenticated:
        return chat_page()
    return redirect(url_for('app.login'))


@app_blueprint.route('/chat')
def chat():
    if current_user.is_authenticated:
        return chat_page()
    return redirect(url_for('app.login'))



from ._base_model import BaseModel


class DummyModel(BaseModel):
    
    def get_response(self, response):
        return response
    
    def invoke(self, messages):
        return '''<div class="thinking_block_signature">thinking_signature</div><div class="thinking_block_content">thinking_content</div> 
        
        
dummy reply
        
<p>This is a paragraph with some <code><p>example</p></code></p>

```css
<style>
p { color: purple!important }
</style>
```

```js
<script>
alert("This is a codeblock example script that should not be executed")
</script>
```

<script>
alert("This is message example script that should not be executed")
</script>

<workspace>
<script>
alert("This is workspace script that should not be executed")
</script>
</workspace>
'''

    async def _async_task(self):
        return 'dummy reply'

    def async_invoke(self, messages):
        import asyncio
        return asyncio.create_task(self._async_task())



import json
import logging
import time
from datetime import datetime, timedelta
from .. import config
from .models import db, User, Conversation, Activity, Subscription, Setting, \
    Message
from .encryption import EncryptionManager
from sqlalchemy import func, delete
from sqlalchemy.orm.exc import NoResultFound

logger = logging.getLogger('sigmund')


class DatabaseManager:
    
    def __init__(self, sigmund, username: str,
                 encryption_key: [str, bytes]=None):
        self._sigmund = sigmund
        self.username = username
        self.encryption_manager = EncryptionManager(encryption_key)
        self.ensure_user_exists()
        logger.info(
            f'initializing database for user {self.username} ({self.user_id})')
        
    def prune_detached_messages(self):
        """Delete all messages that are not attached to a conversation anymore
        due to an old bug in which messages where re-added but never deleted.
        """
        for conversation in Conversation.query.filter_by(
                user_id=self.user_id).all():
            logger.info(f'decrypting {conversation.conversation_id}, {len(conversation.data)} bytes')
            if len(conversation.data) > 20000:
                logger.info('skipping')
                continue
            decrypted_data = self.encryption_manager.decrypt_data(
                conversation.data)
            conversation_data = json.loads(decrypted_data)
            logger.info('getting attached messages')
            attached_message_ids = conversation_data.get('message_history', [])
            attached_message_ids = [
                msg_id for msg_id in attached_message_ids
                if isinstance(msg_id, int)
            ]
            logger.info('getting all messages')
            all_messages = Message.query.filter_by(
                conversation_id=conversation.conversation_id).all()
            for message in all_messages:
                if message.message_id not in attached_message_ids:
                    logger.info(f'deleting {message.message_id}')
                    db.session.delete(message)
        logger.info('committing')
        db.session.commit()
        logger.info('done')
            
    def ensure_user_exists(self):
        try:
            user = User.query.filter_by(username=self.username).one()
            self.user_id = user.user_id
            logger.info(f'user {self.username} exists')
        except NoResultFound:
            # This is a hack to fix a bug in earlier versions of Sigmund without
            # breaking log-in for existing users.
            #
            # If the user id doesn't exist, this may be because the user 
            # previously logged in through Google before the unique ID was 
            # included in the username. Currently, Google usernames look like
            # Name (google)::unique_id. Previously, Google usernames looked like
            # Name (google). If an old Google username exists, we automatically
            # rename the user to the new unique username.
            if '(google)::' in self.username:
                # Strip everything after (google)::
                old_username = self.username.split('(google)::')[0] + '(google)'
                logger.info(
                    f'Checking if {self.username} exists as {old_username}')
                try:                    
                    user = User.query.filter_by(username=old_username).one()
                    self.user_id = user.user_id
                except NoResultFound:
                    # No the user really doesn't exist
                    pass
                else:
                    # Yes, the user has an old-style name, so we need to update
                    # the database so that User.username is set to old username
                    user.username = self.username
                    db.session.commit()
                    return
            # If the user doesn't exist, we create it and start a new 
            # conversation
            logger.info(f'creating new user: {self.username}')
            user = User(username=self.username)
            db.session.add(user)
            db.session.commit()
            self.user_id = user.user_id
            self.new_conversation()
    
    def get_message(self, message_id: int) -> dict:
        # If it's not an int, then it's an old-style mesage
        if not isinstance(message_id, int):
            return message_id
        try:
            # First, check if the message exists in the new Message table
            message = Message.query.filter_by(message_id=message_id).first()
            if message:
                # If it exists, decrypt and return the message
                decrypted_data = self.encryption_manager.decrypt_data(message.data)
                return json.loads(decrypted_data)
            # If the message doesn't exist, something is wrong.
            logger.error(f"Message {message_id} does not exist")
            return {}
        except Exception as e:
            logger.error(f"Error retrieving message {message_id}: {e}")
            return {}
            
    def get_message_history(self, conversation_data):
        """Returns a list of message dicts, excluding empty ones"""
        message_ids = conversation_data.get('message_history', [])
        message_history = []
        for msg_id in message_ids:
            msg = self.get_message(msg_id)
            if msg:
                message_history.append(msg)
        return message_history

    
    def add_message(self, conversation_id: int, message_data: dict) -> int:
        json_data = json.dumps(message_data)
        encrypted_data = self.encryption_manager.encrypt_data(
            json_data.encode('utf-8'))
        message = Message(conversation_id=conversation_id, data=encrypted_data)
        db.session.add(message)
        db.session.commit()
        return message.message_id
    
    def get_active_conversation(self) -> dict:
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=user.active_conversation_id).one()
        except NoResultFound:
            logger.warning(f"No active conversation found for user {self.user_id}")
            return {}    
        decrypted_data = self.encryption_manager.decrypt_data(conversation.data)
        conversation_data = json.loads(decrypted_data)
        conversation_data['message_history'] = self.get_message_history(
            conversation_data)
        return conversation_data

    def set_active_conversation(self, conversation_id: int) -> bool:
        try:
            # We first get the active conversation
            conversation = Conversation.query.filter_by(
                conversation_id=conversation_id, user_id=self.user_id).one()
        except NoResultFound:
            logger.warning(f"Conversation {conversation_id} not found or does "
                           f"not belong to user {self.user_id}")
            return False
        # And then update the current time for that conversation so that
        # it ends on top
        decrypted_data = self.encryption_manager.decrypt_data(
            conversation.data)
        conversation_data = json.loads(decrypted_data)
        conversation_data['last_updated'] = time.time()
        json_data = json.dumps(conversation_data)
        encrypted_data = self.encryption_manager.encrypt_data(
            json_data.encode('utf-8'))
        conversation.data = encrypted_data
        # Finally we change the active conversation for the user
        user = User.query.filter_by(user_id=self.user_id).one()
        user.active_conversation_id = conversation.conversation_id
        db.session.commit()
        return True
    
    def update_active_conversation(self, conversation_data: dict) -> bool:
        """Updates the active conversation. The messages are saved in a 
        separate table as separate entries. However, all messages are replaced
        when the conversation is updated. That the conversation doesn't grow
        when a new message is added, but rather is reconstructed.
        """
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=user.active_conversation_id
            ).one()
        except NoResultFound:
            logger.warning(
                f"No active conversation to update for user {self.user_id}")
            return False            
        logger.info(f'updating conversation {user.active_conversation_id}')
        logger.info('deleting old messages')
        # Batch delete all messages linked to this conversation, because we
        # will recreate new messages
        db.session.execute(
            delete(Message).where(
                Message.conversation_id == conversation.conversation_id)
        )
        logger.info('adding new messages')
        # Extract new messages and store them separately
        message_history = conversation_data.pop('message_history', [])
        message_ids = []
        for message in message_history:
            message_id = self.add_message(conversation.conversation_id,
                                          message)
            message_ids.append(message_id)
        conversation_data['message_history'] = message_ids
        conversation_data['last_updated'] = time.time()
        json_data = json.dumps(conversation_data)
        encrypted_data = self.encryption_manager.encrypt_data(
            json_data.encode('utf-8'))
        conversation.data = encrypted_data
        logger.info('committing conversation')
        db.session.commit()
        logger.info('done')
        return True

    def list_conversations(self, query=None) -> dict:
        conversations = {}
        user = User.query.filter_by(user_id=self.user_id).one()
        for conversation in \
                Conversation.query.filter_by(user_id=self.user_id).all():
            try:
                data = json.loads(self.encryption_manager.decrypt_data(
                    conversation.data))
                # Don't list empty conversations except for the active one
                if len(data.get('message_history', [])) < 2 and \
                        user.active_conversation_id != conversation.conversation_id:
                    continue
                title = data.get('title', 'Untitled conversation')
                if query is not None:
                    # First check title match
                    if query.lower() not in title.lower():
                        # If the title doesn't match, then we do a full-text
                        # search on the message history
                        match = False
                        for message in self.get_message_history(data):
                            message_text = message[1]
                            if query.lower() in message_text.lower():
                                match = True
                                break
                        if not match:
                            continue
                conversations[conversation.conversation_id] = (
                     title, data.get('last_updated', time.time()))
            except Exception as e:
                logger.error(f"Error decrypting conversation data: {e}")
        return conversations
    
    def export_conversations(self) -> dict:
        conversations = []
        for conversation in \
                Conversation.query.filter_by(user_id=self.user_id).all():
            try:
                decrypted_data = self.encryption_manager.decrypt_data(
                    conversation.data)
                conversation_data = json.loads(decrypted_data)
                message_history = self.get_message_history(conversation_data)
            except Exception as e:
                logger.error(f"Error decrypting conversation data: {e}")
            else:
                conversation_data['message_history'] = message_history
                conversations.append(conversation_data)
        return conversations    

    def new_conversation(self) -> bool:
        try:
            conversation_data = {
                "title": "New conversation",
                "condensed_text": None,
                "message_history": [],
                "message_metadata": [],
                "last_updated": time.time()
            }
            json_data = json.dumps(conversation_data)
            encrypted_data = self.encryption_manager.encrypt_data(
                json_data.encode('utf-8'))
            conversation = Conversation(user_id=self.user_id,
                                        data=encrypted_data)
            db.session.add(conversation)
            db.session.commit()
            self.set_active_conversation(conversation.conversation_id)
            return True
        except Exception as e:
            logger.error(f"Error creating new conversation: {e}")
            return False
    
    def delete_conversation(self, conversation_id: int) -> bool:
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=conversation_id, user_id=self.user_id).one()
            if user.active_conversation_id == conversation.conversation_id:
                logger.warning(
                    f"Cannot delete active conversation {conversation_id} for "
                    f"user {self.user_id}")
                return False

            # Delete associated messages
            Message.query.filter_by(conversation_id=conversation_id).delete()

            # Delete the conversation
            db.session.delete(conversation)
            db.session.commit()
            return True
        except NoResultFound:
            logger.warning(f"Conversation {conversation_id} not found or does "
                           f"not belong to user {self.user_id}")
            return False

    def add_activity(self, tokens_consumed: int):
        """Adds the number of tokens consumed using the current time."""
        new_activity = Activity(user_id=self.user_id, time=datetime.utcnow(),
                                tokens_consumed=tokens_consumed)
        db.session.add(new_activity)
        db.session.commit()

    def get_activity(self, after_time=None) -> int:
        """Retrieves the total number of tokens consumed since a particular 
        time, using the past hour as a default if no after_time is provided.
        """
        if after_time is None:
            after_time = datetime.utcnow() - timedelta(hours=1)

        total_tokens = db.session.query(func.sum(Activity.tokens_consumed)) \
            .filter(Activity.user_id == self.user_id) \
            .filter(Activity.time >= after_time) \
            .scalar()
        return total_tokens if total_tokens is not None else 0

    def update_subscription(self, stripe_customer_id: str,
                            stripe_subscription_id: str, from_date=None,
                            to_date=None):
        """Updates a subscription. By default the subscriptions starts 
        immediately and ends exactly one month from the current time. A user
        has only one subscription, which means that a subscription should be
        added if it doesn't exist, and updated if it already exists.
        """
        now = datetime.utcnow()
        from_date = from_date or now
        to_date = to_date or now + timedelta(days=config.subscription_length)
        subscription = Subscription.query.filter_by(
            user_id=self.user_id).first()
        if subscription:
            subscription.from_date = from_date
            subscription.to_date = to_date
        else:
            subscription = Subscription(
                user_id=self.user_id, from_date=from_date, to_date=to_date,
                stripe_customer_id=stripe_customer_id,
                stripe_subscription_id=stripe_subscription_id)
            db.session.add(subscription)
        db.session.commit()

    def check_subscription(self) -> bool:
        """Returns a bool indicating whether the user is currently subscribed.
        """
        now = datetime.utcnow()
        subscription = Subscription.query.filter_by(
            user_id=self.user_id).first()
        return subscription and \
            subscription.from_date <= now < subscription.to_date

    def cancel_subscription(self):
        """Sets the to_date of the subscription to the current time if the user
        is currently subscribed.
        """
        if not self.check_subscription():
            return
        subscription = Subscription.query.filter_by(
            user_id=self.user_id).first()
        subscription.to_date = datetime.utcnow()
        db.session.commit()

    def get_stripe_customer_id(self) -> str:
        subscription_record = Subscription.query.filter(
            Subscription.user_id == self.user_id
        ).order_by(Subscription.from_date.desc()).first()
        if subscription_record:
            return subscription_record.stripe_customer_id
        return None

    @staticmethod
    def from_stripe_customer_id(stripe_customer_id: str) -> str:
        subscription_record = Subscription.query.filter(
            Subscription.stripe_customer_id == stripe_customer_id
        ).order_by(Subscription.from_date.desc()).first()
        if not subscription_record:
            return None
        user_record = User.query.filter(
            User.user_id == subscription_record.user_id
        ).first()
        if not user_record:
            return None
        return DatabaseManager(None, username=user_record.username)

    def get_setting(self, key: str) -> str:
        """Retrieve a setting value for the current user, which is available
        as self.user_id. If the setting does not exist, return the default
        value as specified in the config or None if no default has been 
        specified.
        """
        setting = db.session.query(Setting).filter_by(
            user_id=self.user_id, key=key).first()
        return setting.value if setting \
            else config.settings_default.get(key, None)

    def set_setting(self, key: str, value: str):
        """Set a setting to specified value for the current user. If the
        setting already exists, overwrite it.
        """
        setting = Setting.query.filter_by(
            user_id=self.user_id, key=key).first()
        if setting:
            setting.value = value
        else:
            new_setting = Setting(user_id=self.user_id, key=key,
                                  value=value)
            db.session.add(new_setting)
        db.session.commit()

    def list_settings(self) -> dict:
        """Returns a dict of all settings for the current user."""
        settings = {}
        setting_objs = db.session.query(Setting).filter_by(
            user_id=self.user_id).all()
        for setting in setting_objs:
            settings[setting.key] = setting.value
        return settings



from . import BaseTool
import logging
import requests
logger = logging.getLogger('sigmund')


class execute_code(BaseTool):
    """Add Python or R code to the workspace and execute it. Do not use this tool unless code should actually be executed. When executing Python, you can only use modules from the Python standard library."""
    
    arguments = {
        'language': {
            'type': 'string',
            'description': 'The programming language to use',
            'enum': ['r', 'python']
        },
        'code': {
            'type': 'string',
            'description': 'The code to execute. Use print() to print to the standard output.'
        }
    }
    required_arguments = ['language', 'code']
    
    def __call__(self, language, code):
        logger.info(f'executing {language} code: {code}')
        url = "https://emkc.org/api/v2/piston/execute"
        language_aliases = {'python': 'python',
                            'r': 'r'}
        language = language_aliases[language.lower()]
        language_versions = {'python': '3.10', 'r': '4.1.1'}
        language_files = {'python': 'main.py', 'r': 'main.R'}
        data = {
            "language": language,
            "version": language_versions[language],
            "files": [{"name": language_files[language], "content": code}],
            "stdin": "",
            "args": [],
            "compile_timeout": 10000,
            "run_timeout": 3000,
            "compile_memory_limit": -1,
            "run_memory_limit": -1
        }
        response = requests.post(url, json=data)
        if response.status_code == 200:
            response_data = response.json()
            result = response_data.get("run", {}).get("output", "").strip()
            logger.info(f'result: {result}')
            message = f'''I added code to the workspace, executed it, and received the following output:

```
{result}
```

<workspace language="{language}">
{code}
</workspace>
'''
            return message, result, language, True
        logger.error(f"Error: {response.status_code} with message: {response.content}")
        return 'Failed to execute code', code, language, True



import os
os.environ['USE_FLASK_SQLALCHEMY'] = '1'
from flask import Flask, Config, request
from flask_login import LoginManager
from flask_migrate import Migrate
from flask_session import Session
from . import config
from .redis_client import redis_client
from .routes import api_blueprint, app_blueprint, User, subscribe_blueprint, \
    google_login_blueprint, public_blueprint
from . import utils
from .database.models import db
import logging
logger = logging.getLogger('sigmund')
logging.basicConfig(level=logging.INFO, force=True)


class SigmundConfig(Config):
    SECRET_KEY = config.flask_secret_key
    SQLALCHEMY_DATABASE_URI = 'sqlite:///sigmund.db'
    SESSION_TYPE = 'redis'
    SESSION_PERMANENT = False
    SESSION_USE_SIGNER = True
    SESSION_REDIS = redis_client


def create_app(config_class=SigmundConfig):
    app = Flask(__name__, static_url_path='/static')
    app.config.from_object(config_class)
    # Initialize Flask-Session to activate Redis for session management
    Session(app)
    app.register_blueprint(app_blueprint)
    app.register_blueprint(api_blueprint, url_prefix='/api')
    app.register_blueprint(subscribe_blueprint, url_prefix='/subscribe')
    app.register_blueprint(google_login_blueprint, url_prefix='/google_login')
    app.register_blueprint(public_blueprint, url_prefix='/public')
    # Initialize the databasea
    db.init_app(app)
    with app.app_context():
        db.create_all()
    # Initialize login manager
    login_manager = LoginManager()
    # Set up database migration
    migrate = Migrate(app, db)
    @login_manager.user_loader
    def load_user(user_id):
        return User(user_id)
        
    login_manager.init_app(app)

    @app.after_request
    def log_request(response):
        user_agent = request.headers.get('User-Agent')
        if user_agent is not None and 'bot' not in user_agent.lower():
            logger.info(f'request: {request.full_path}')
        return response
        
    @app.errorhandler(Exception)
    def handle_exception(e):
        logger.error(e)
        return utils.render('error.html'), 500

    return app



import html
import logging
import textwrap
from bs4 import BeautifulSoup
import time
import re
from flask import render_template, render_template_string
import markdown
from markdown.extensions.fenced_code import FencedCodeExtension
from markdown.extensions.codehilite import CodeHiliteExtension
from markdown.extensions.toc import TocExtension
from markdown.extensions.attr_list import AttrListExtension
from markdown.extensions.md_in_html import MarkdownInHtmlExtension
from langchain.schema import HumanMessage
from . import config
from . import __version__
logger = logging.getLogger('sigmund')


def md(text):
    return markdown.markdown(text,
                             extensions=[FencedCodeExtension(),
                                         CodeHiliteExtension(),
                                         TocExtension(),
                                         AttrListExtension(),
                                         MarkdownInHtmlExtension()])


def clean(text, escape_html=True, render=True):
    if render:
        text = render_template_string(text)
    if escape_html:
        text = html.escape(text)
    return text
    

def render(path, **kwargs):
    return render_template(
        path,
        ai_name=config.ai_name,
        page_title=config.page_title,
        server_url=config.server_url,
        max_message_length=config.max_message_length,
        version=__version__,
        **kwargs)


def prepare_messages(messages, allow_ai_first=True, allow_ai_last=True,
                     merge_consecutive=False, merge_separator='\n'):
    """Takes a list of Message objects formats them to meet the requirements
    of specific models.
    
    Parameters
    ----------
    messages : list
        A list of messages, where each message object has a type property that
        is either 'system', 'ai', or ;'human'
    allow_ai_first : bool, optional
        Indicates whether the first message (after the system message) can be
        an AI message. If not, then it is removed.
    allow_ai_last : bool, optional
        Indicates whether the last message (after the system message) can be
        an AI message. If not, then a human message is appended after it.
    merge_consecutive : bool, optional
        Indicates whether multiple messages of the same class should be merged
        into a single message.
    merge_separator : str, optional
        If messages are merged, the separator that is used to join their 
        content.
        
    Returns
    -------
    list
    """
    if not isinstance(messages, list):
        return messages
    if not messages:
        return []
    # Check whether first message after the system message is an AI message, 
    # and remove this if not allowed
    if not allow_ai_first and messages[1].type == 'ai':
        logger.info('removing first assistant mesage')
        messages.pop(1)        
    # Merge consecutive messages of the same class totether
    if len(messages) > 1 and merge_consecutive:
        merged_messages = []
        current_message = messages[0]
        for next_message in messages[1:]:
            if current_message.type == next_message.type:
                logger.info('merging consecutive messages')
                current_message.content += merge_separator + next_message.content
            else:
                merged_messages.append(current_message)
                current_message = next_message
        merged_messages.append(current_message)
        messages = merged_messages
    # Make sure the last message is human if this is required
    if not allow_ai_last and messages[-1].type == 'ai':
        logger.info('adding continue message')
        messages.append(HumanMessage(content='Please continue!'))
    return messages


def extract_workspace(txt: str) -> tuple:
    """Takes a string of text and extracts workspace content from it. There are
    a few ways in which this can occur:
    
    - If the text contains a workspace indicated like this:
      <workspace language="language">
      your workspace content
      </workspace>`
      Then language and the content should be extracted and returned as the 
      second value of the tuple. If no language is specified, it defaults to
      'markdown'. The workspace tags and content should be stripped from txt.
    - If the text doesn't contain an explicit workspace, but does contain 
      markdown code blocks like this:
      ```language 
      code here
      ```
      Then this code should be extracted as the workspace content and language,
      again falling back markdown if no language is provided. The markdown
      code blocks do not have to be stripped from txt.
      
    Return txt, workspace_content, language
    """
    open_pattern = r'^<workspace(?: language="(.+?)")?>'
    if re.search(open_pattern, txt, re.DOTALL | re.MULTILINE) \
            and '</workspace>' not in txt:
        logger.warning('workspace closing tag appears to be missing')
        txt += '\n</workspace>'
    # Checks for <workspace> tags
    pattern = r'^<workspace(?: language="(.+?)")?>(.*?)^</workspace>'
    match = re.search(pattern, txt, re.DOTALL | re.MULTILINE)
    if match:
        language = match.group(1) if match.group(1) else 'markdown'
        content = match.group(2).strip()
        # Only replace the first occurrence (the one we extracted)
        text_without_workspace = re.sub(pattern, "", txt, count=1,
                                        flags=re.DOTALL | re.MULTILINE).strip()
        
        # Convert any remaining workspace tags to markdown code blocks
        def replace_workspace_with_codeblock(m):
            lang = m.group(1) if m.group(1) else 'markdown'
            code_content = m.group(2).strip()
            return f'```{lang}\n{code_content}\n```'
        
        text_without_workspace = re.sub(pattern, replace_workspace_with_codeblock, 
                                      text_without_workspace, 
                                      flags=re.DOTALL | re.MULTILINE)
        
        # Empty messages can cause issues, so if stripping the workspace 
        # results in an empty message, we add a placeholder.
        if not text_without_workspace:
            text_without_workspace = 'I added content to the workspace!'
        return text_without_workspace, content, language
    # Checks for ``` code blocks
    pattern_codeblock = r'^```(?:([a-z]+))?\n(.*?)^```'
    match_codeblock = re.search(pattern_codeblock, txt, re.DOTALL | re.MULTILINE)
    if match_codeblock:
        language = match_codeblock.group(1) if match_codeblock.group(1) else 'markdown'
        content = match_codeblock.group(2).strip()
        if len(content.splitlines()) > 2:
            return txt, content, language
    # Simply returns txt if no workspace is detected
    return txt, None, None

def remove_masked_elements(content):
    # This pattern matches:
    #   1) an opening tag <tag ...>
    #   2) that includes class="...mask..."
    #   3) everything inside (including nested tags)
    #   4) up to the matching </tag>
    #
    # It does this by capturing the tag name in group 1, so we can search for
    # the matching </tag> later, and also checking for "mask" in the class attribute.
    pattern = re.compile(
        r'<([a-zA-Z0-9]+)([^>]*)\bclass\s*=\s*[\'"]([^"\']*\bmask\b[^"\']*)[\'"]([^>]*)>(.*?)</\1>',
        re.DOTALL
    )
    return re.sub(pattern, '', content)


def current_datetime():
    return time.strftime('%a %d %b %Y %H:%M')



import json
import logging
from .library import Library
from . import config
logger = logging.getLogger('sigmund')


class Documentation:
    
    def __init__(self, sigmund):
        self._sigmund = sigmund
        self._documents = []
        self.poor_match = False
        self._library = Library(
            persist_directory=config.search_persist_directory,
            embedding_provider=config.search_embedding_provider,
            embedding_model=config.search_embedding_model)
        self._collections = {
            c for c in config.search_collections
            if self._sigmund.database.get_setting(f'collection_{c}') == 'true'
        }
        logger.info(f'using collections: {self._collections}')
        
    @property
    def enabled(self):
        return self._collections
        
    def _doc_to_str(self, doc):
        s = '<document>\n'
        if doc.get('title'):
            s += f'# {doc["title"]}\n\n'
        if doc.get('url'):
            s += f'Source: {doc["url"]}\n\n'
        return s + doc['content'] + '\n</document>'
        
    def __str__(self):
        if not self._documents:
            return ''
        return '\n\n'.join(self._doc_to_str(doc) for doc in self._documents)
        
    def to_json(self):
        return json.dumps([{'page_content': doc['content'],
                            'url': doc.get('url')}
                           for doc in self])
        
    def __iter__(self):
        return (doc for doc in self._documents)
    
    def __len__(self):
        return sum(len(doc['content']) for doc in self._documents)
        
    def __contains__(self, doc):
        return doc in self._documents
    
    def prompt(self):        
        return f'''# Documentation

You have retrieved the following documentation to answer the user's question:

<documentation>
{str(self)}
</documentation>'''
    
    def append(self, doc):
        if any(doc['content'] == d['content'] for d in self):
            return
        self._documents.append(doc)        
    
    def clear(self):
        logger.info('clearing documentation')
        self._documents = []
        
    def search(self, query, fallback=False, foundation=True, howtos=True,
               max_distance=None, max_distance_fallback=None, k=None):
        """First, we separately search for regular and howto documents, because
        these tend not to be fairly comparable (howtos tend to always win).
        Finally, we insert foundation documents that match the topic of the
        search hits.
        """
        if not self._collections:
            logger.info('library search disabled')
            return
        if k is None:
            k = config.search_docs_max
        if max_distance is None:
            max_distance = config.search_max_distance
        if max_distance_fallback is None:
            max_distance_fallback = config.search_max_distance_fallback
        if fallback:
            max_distance = max_distance_fallback
        regular_results = self._library.search(
            query, foundation=False, howto=False, k=k,
            max_distance=max_distance, collection=self._collections)        
        logger.info(f'found {len(regular_results)} regular results')
        if howtos:
            howto_results = self._library.search(
                query, foundation=False, howto=True, k=k,
                max_distance=max_distance, collection=self._collections)
            logger.info(f'found {len(howto_results)} howto results')
        else:
            howto_results = []
        results = regular_results + howto_results
        if foundation:
            foundation_results = self._library.retrieve_foundation_documents(results)
            logger.info(f'found {len(foundation_results)} foundation documents')
        else:
            foundation_results = []
        self._documents = foundation_results + results
        if not fallback and not self._documents:
            self.poor_match = True
            logger.warning('no results found, retrying with higher threshold')
            self.search(query, fallback=True, foundation=foundation, 
                        howtos=howtos, max_distance=max_distance,
                        max_distance_fallback=max_distance_fallback, k=k)



from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField, SubmitField
from wtforms.validators import DataRequired


class LoginForm(FlaskForm):
    username = StringField('Username (from forum.cogsci.nl)',
                           validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired()])
    submit = SubmitField('Log In')
